import json
import math
import os
from typing import List, Tuple

import cv2
import numpy as np
import onnxruntime
from lanyocr.lanyocr_utils import download_model
from lanyocr.text_recognizer import LanyOcrRecognizer
from lanyocr.text_recognizer.mmocr import MMOCR


class MMOCR_Satrn_Sm(MMOCR):
    DICT90 = tuple(
        "0123456789abcdefghijklmnopqrstuvwxyz"
        "ABCDEFGHIJKLMNOPQRSTUVWXYZ!\"#$%&'()"
        "*+,-./:;<=>?@[\\]_`~"
    )

    def __init__(self, use_gpu: bool = True) -> None:
        model_path = download_model("MMOCR_SATRN_SM.onnx")

        self.idx2char = list(self.DICT90)

        self.start_end_token = "<BOS/EOS>"
        self.unknown_token = "<UKN>"
        self.padding_token = "<PAD>"

        self.idx2char.append(self.unknown_token)
        self.unknown_idx = len(self.idx2char) - 1

        self.idx2char.append(self.start_end_token)
        self.start_idx = len(self.idx2char) - 1
        self.end_idx = len(self.idx2char) - 1

        self.idx2char.append(self.padding_token)
        padding_idx = len(self.idx2char) - 1

        self.ignore_indexes = [padding_idx]

        super().__init__(use_gpu, model_path, 100, 32)

        self.set_max_batch_size(1)

    def normalize_img(self, bgr_img):
        img = cv2.resize(bgr_img, (self.model_w, self.model_h)).astype(np.float32)[
            :, :, ::-1
        ]
        img /= 255

        img[:, :, 0] -= 0.485
        img[:, :, 1] -= 0.456
        img[:, :, 2] -= 0.406

        img[:, :, 0] /= 0.229
        img[:, :, 1] /= 0.224
        img[:, :, 2] /= 0.225

        img = np.transpose(img, [2, 0, 1])

        return img

    def decode(self, pred) -> Tuple[str, float]:
        result = np.argmax(pred, axis=-1)

        text = ""
        score = 1.0
        for idx, scores in zip(result, pred):
            if idx in self.ignore_indexes:
                continue

            if idx == self.end_idx:
                break

            text += self.idx2char[idx]
            score *= scores[idx]

        return (text, score)

    def set_max_batch_size(self, batch_size: int):
        # this model does not support batching
        self.max_batch_size = 1



import json
import math
import os
from typing import List, Tuple

import cv2
import numpy as np
import onnxruntime
from lanyocr.lanyocr_utils import download_model
from lanyocr.text_recognizer import LanyOcrRecognizer


class MMOCR(LanyOcrRecognizer):
    def __init__(
        self,
        use_gpu: bool = True,
        model_path: str = "",
        model_w: int = 100,
        model_h: int = 32,
    ) -> None:
        super().__init__(use_gpu)

        self.model_w = model_w
        self.model_h = model_h
        self.model_path = model_path

        # load model
        providers = ["CPUExecutionProvider"]
        if self.use_gpu:
            providers = ["CUDAExecutionProvider"]

        opts = onnxruntime.SessionOptions()
        opts.graph_optimization_level = (
            onnxruntime.GraphOptimizationLevel.ORT_ENABLE_EXTENDED
        )

        self.session = onnxruntime.InferenceSession(
            self.model_path, sess_options=opts, providers=providers
        )

    def infer(self, bgr_img) -> Tuple[str, float]:
        norm_img = self.normalize_img(bgr_img)

        pred = self.session.run(None, {"input": [norm_img]})[0][0]

        return self.decode(pred)

    def infer_batch(self, bgr_images) -> List[Tuple[str, float]]:
        batch_results = []

        normed_imgs = [self.normalize_img(img) for img in bgr_images]

        for idx in range(0, len(normed_imgs), self.max_batch_size):
            batch_preds = self.session.run(
                None, {"input": normed_imgs[idx : idx + self.max_batch_size]}
            )[0]

            for pred in batch_preds:
                result = self.decode(pred)
                batch_results.append(result)

        return batch_results

    def get_model_height(self) -> int:
        return self.model_h

    def get_model_width(self) -> int:
        return self.model_w

    def normalize_img(self, bgr_img):
        raise NotImplementedError

    def decode(self, pred) -> Tuple[str, float]:
        raise NotImplementedError



import os
from glob import glob

from lanyocr.benchmarker import LanyBenchmarker


class LanyBenchmarkerICDAR2015(LanyBenchmarker):
    def load_dataset(self):
        image_paths_dir = os.path.join(self.dataset_path, "images")
        gt_paths_dir = os.path.join(self.dataset_path, "validation")

        image_paths = glob(image_paths_dir + "/*")
        gt_paths = glob(gt_paths_dir + "/*")

        for image_path in image_paths:
            image_name = os.path.basename(image_path)
            image_id = int(image_name.split(".")[0].split("_")[-1])
            self.id2imagePath[image_id] = image_path

        for gt_path in gt_paths:
            gt_name = os.path.basename(gt_path)
            gt_id = int(gt_name.split(".")[0].split("_")[-1])
            self.id2gts[gt_id] = (
                open(gt_path, "r", encoding="utf-8-sig").read().encode()
            )

        assert (
            len(self.id2imagePath) != 0
        ), "Make sure dataset_path contains images and validation directories"



import json
import os
from typing import List
from typing import Tuple

import cv2
import numpy as np

from lanyocr.lanyocr_utils import download_model
from lanyocr.text_recognizer.paddleocr_base import PaddleOcrBase


class PaddleOcrEnPPOCRV3(PaddleOcrBase):
    def __init__(self, use_gpu: bool = True) -> None:
        print("Recognizer: PaddleOcrChv3")

        model_h = 48
        model_w = 320

        cur_dir = os.path.dirname(os.path.realpath(__file__))

        model_ignored_tokens = [0]
        accepted_characters = json.load(
            open(os.path.join(cur_dir, "dicts/paddleocr_en_dict.json"))
        )
        # accepted_characters = []
        model_characters = json.load(
            open(os.path.join(cur_dir, "dicts/paddleocr_en_dict.json"))
        )
        model_path = download_model("lanyocr-en-ppocrv3.onnx")

        assert os.path.exists(model_path)

        super().__init__(
            use_gpu,
            model_ignored_tokens,
            model_characters,
            accepted_characters,
            model_path,
            model_w,
            model_h,
        )

        self.set_max_batch_size(1)

    def normalize_img(self, bgr_img):
        h, w = bgr_img.shape[:2]

        # scale based on h
        ratio = float(self.model_h) / h
        resized_w = int(w * ratio)

        resized_image = cv2.resize(bgr_img, (resized_w, self.model_h))

        resized_image = resized_image.astype("float32")
        resized_image = resized_image.transpose((2, 0, 1)) / 255.0
        resized_image -= 0.5
        resized_image /= 0.5

        return resized_image



import math
import os
from typing import List

import cv2
import numpy as np

from lanyocr.angle_classifier import LanyOcrAngleClassifierFactory
from lanyocr.lanyocr_utils import LanyOcrResult
from lanyocr.lanyocr_utils import LanyOcrRRect
from lanyocr.lanyocr_utils import LanyOcrTextLine
from lanyocr.lanyocr_utils import crop_rrect
from lanyocr.lanyocr_utils import distance
from lanyocr.lanyocr_utils import resize_img_to_height
from lanyocr.lanyocr_utils import resize_img_to_width
from lanyocr.lanyocr_utils import rotate_image
from lanyocr.lanyocr_utils import rotate_image_at_center
from lanyocr.text_detector import LanyOcrDetectorFactory
from lanyocr.text_merger import LanyOcrMergerFactory
from lanyocr.text_recognizer import LanyOcrRecognizerFactory


class LanyOcr:
    def __init__(
        self,
        detector_name: str = "paddleocr_en_ppocr_v3",
        recognizer_name: str = "paddleocr_en_ppocr_v3",
        angle_classifier_name: str = "paddleocr_mobile",
        merger_name: str = "lanyocr_nomerger",
        merge_boxes_inference: bool = False,
        merge_rotated_boxes: bool = True,
        merge_vertical_boxes: bool = False,
        use_gpu: bool = False,
        debug: bool = False,
    ) -> None:
        if merge_boxes_inference and recognizer_name in [
            "mmocr_satrn",
            "mmocr_satrn_sm",
            "paddleocr_en_ppocr_v3",
        ]:
            merge_boxes_inference = False
            if recognizer_name == "paddleocr_en_ppocr_v3":
                print(
                    f"Disabled merge_boxes_inference because {recognizer_name} has dynamic input width."
                )
            else:
                print(
                    f"Disabled merge_boxes_inference because {recognizer_name} could not recognize space character."
                )

        self.detector = LanyOcrDetectorFactory.create(detector_name, use_gpu=use_gpu)
        self.recognizer = LanyOcrRecognizerFactory.create(
            recognizer_name, use_gpu=use_gpu
        )
        self.angle_classifier = LanyOcrAngleClassifierFactory.create(
            angle_classifier_name, use_gpu=use_gpu
        )
        self.merger = LanyOcrMergerFactory.create(merger_name)

        self.rotate_90degrees_thresh = 1.5
        self.spacing = 10
        self.merge_boxes_inference = merge_boxes_inference
        self.merge_rotated_boxes = merge_rotated_boxes
        self.merge_vertical_boxes = merge_vertical_boxes

        self.debug: bool = debug
        self.debug_dir = "./debug"

        if self.debug:
            import shutil

            if os.path.exists(self.debug_dir):
                shutil.rmtree(self.debug_dir)
            os.makedirs(self.debug_dir)

    def infer_from_file(self, image_path: str) -> List[LanyOcrResult]:
        image = cv2.imread(image_path)
        return self.infer(image)

    def infer(self, brg_image) -> List[LanyOcrResult]:
        ocr_results: List[LanyOcrResult] = []

        det_rrects = self.detector.infer(brg_image)
        det_lines = self.merger.merge_to_lines(
            det_rrects, self.merge_rotated_boxes, self.merge_vertical_boxes
        )

        original_img = np.array(brg_image)

        for line_idx, line in enumerate(det_lines):
            result = self._infer_line(original_img, line, line_idx)
            if result.text != "":
                ocr_results.append(result)

            if self.debug:
                print(f'Line {line_idx} - Text: "{result.text}" - Score: {result.prob}')

        return ocr_results

    def visualize(
        self,
        image,
        results: List[LanyOcrResult],
        visualize_sub_boxes: bool = False,
        output_path: str = "outputs/output.jpg",
        font_scale: float = 0.5,
    ):
        print("Visualizing results...")

        COLORS = [
            # (0, 0, 0),
            (255, 0, 0),
            (0, 255, 0),
            (0, 0, 255),
            (255, 255, 0),
            # (0, 255, 255),
        ]

        if isinstance(image, str):
            image = cv2.imread(image)

        vis_img = np.ones(image.shape, dtype=np.uint8) * 255

        for idx, result in enumerate(results):
            line_rrect = result.line.get_rrect()
            line_angle = line_rrect[2]

            if abs(line_angle) > 89:
                line_angle = 90

            if line_angle == 0 and line_rrect[1][0] * 1.3 < line_rrect[1][1]:
                line_angle = 90
            elif line_angle == 90 and line_rrect[1][1] > 1.3 * line_rrect[1][0]:
                line_angle = 0

            if line_angle not in [0, 90]:
                points = cv2.boxPoints(line_rrect)
                points = np.array(points).reshape((1, -1, 2))
                sorted_cnts = points[0][points[0][:, 0].argsort()[::-1]]
                lowestPointIdx = np.argmax(sorted_cnts[:, 1])
                num_pnt_on_left = points.shape[1] - lowestPointIdx - 1
                if num_pnt_on_left < 2:
                    if line_angle < 45:
                        line_angle = -line_angle
                    else:
                        line_angle = -(90 - line_angle)
                else:
                    line_angle = line_angle

            points = cv2.boxPoints(line_rrect)
            box = np.int0(np.array(points))
            center_x = np.sum(box[:, 0]) // 4
            center_y = np.sum(box[:, 1]) // 4

            vis_img = cv2.drawContours(vis_img, [box], 0, COLORS[idx % len(COLORS)], 2)
            image = cv2.drawContours(image, [box], 0, COLORS[idx % len(COLORS)], 2)

            t_size = cv2.getTextSize(result.text, 0, fontScale=font_scale, thickness=1)[
                0
            ]

            center_x = int(center_x)
            center_y = int(center_y)

            padding_x = max(center_x, image.shape[1] - center_x)
            padding_y = max(center_y, image.shape[0] - center_y)
            padding_size = max(padding_x, padding_y) + 1

            vis_img = cv2.copyMakeBorder(
                vis_img,
                padding_size,
                padding_size,
                padding_size,
                padding_size,
                cv2.BORDER_CONSTANT,
            )

            vis_img = rotate_image_at_center(
                vis_img,
                (center_x + padding_size, center_y + padding_size),
                line_angle,
            )

            vis_img = cv2.putText(
                vis_img,
                f"{result.text}",
                (
                    int(center_x + padding_size - t_size[0] // 2),
                    int(center_y + padding_size + t_size[1] // 2),
                ),
                cv2.FONT_HERSHEY_SIMPLEX,
                font_scale,
                (0, 0, 255),
                1,
                cv2.LINE_AA,
            )

            image = cv2.putText(
                image,
                f"{idx}",
                (
                    int(center_x),
                    int(center_y),
                ),
                cv2.FONT_HERSHEY_SIMPLEX,
                0.7,
                (0, 0, 255),
                1,
                cv2.LINE_AA,
            )

            vis_img = rotate_image_at_center(
                vis_img,
                (center_x + padding_size, center_y + padding_size),
                -line_angle,
            )

            vis_img = vis_img[
                padding_size : padding_size + image.shape[0],
                padding_size : padding_size + image.shape[1],
                :,
            ]

            if visualize_sub_boxes:
                for sub_rrect in result.line.sub_rrects:
                    points = sub_rrect.points
                    box = np.int0(np.array(points))
                    vis_img = cv2.drawContours(
                        vis_img, [box], 0, COLORS[(idx + 1) % len(COLORS)], 1
                    )

        if image.shape[1] >= 1.5 * image.shape[0]:
            vis_img = np.concatenate([image, vis_img], axis=0)
        else:
            vis_img = np.concatenate([image, vis_img], axis=1)

        cv2.imwrite(output_path, vis_img)

        print(f"Stored output image in {output_path}")

    def _infer_line(
        self, original_img, line: LanyOcrTextLine, line_idx: int
    ) -> LanyOcrResult:
        text = ""
        prob = 1.0

        line_rrect = line.get_rrect()

        line_img = crop_rrect(np.array(original_img), line_rrect)

        line_h, line_w = line_img.shape[:2]
        if line_h > self.rotate_90degrees_thresh * line_w:
            line_img = cv2.rotate(line_img, cv2.ROTATE_90_CLOCKWISE)

        # determine if a text is 180 degrees flipped or not
        line_flipped = self._is_line_flipped(line_img, 0.925)
        if line_flipped:
            line_img = cv2.rotate(line_img, cv2.ROTATE_180)

        merged_sub_rrect_list: List[List[LanyOcrRRect]] = []
        merged_imgs: List = []

        merged_img = (
            np.ones(
                shape=(
                    self.recognizer.get_model_height(),
                    self.recognizer.get_model_width(),
                    3,
                ),
                dtype=np.uint8,
            )
            * 255
        )
        s = 0

        for sub_rrect_idx, sub_rrect in enumerate(line.sub_rrects):
            if len(merged_sub_rrect_list) == 0:
                merged_sub_rrect_list.append([])

            dst_img = self._crop_and_rotate_rrect_horizontal(
                line_img, line_rrect, sub_rrect, line_flipped
            )

            if not self.merge_boxes_inference:
                # inference on each individual box for best accuracy
                merged_sub_rrect_list[-1].append(sub_rrect)
                merged_imgs.append(dst_img)
                if sub_rrect_idx < len(line.sub_rrects) - 1:
                    merged_sub_rrect_list.append([])
                continue

            resized_dst_img = resize_img_to_height(dst_img, 32)
            h, w = resized_dst_img.shape[:2]

            if s + w >= self.recognizer.get_model_width() - 5:
                # do not merge sub-rrect into single image anymore
                merged_img[:, s + 1 :, :] = 127
                merged_imgs.append(merged_img)
                # initialize new merge
                merged_img = (
                    np.ones(
                        shape=(
                            self.recognizer.get_model_height(),
                            self.recognizer.get_model_width(),
                            3,
                        ),
                        dtype=np.uint8,
                    )
                    * 255
                )
                s = 0
                merged_sub_rrect_list.append([])

            merged_sub_rrect_list[-1].append(sub_rrect)

            if w <= self.recognizer.get_model_width():
                merged_img[:, s : s + w, :] = resized_dst_img
            else:
                # very long sub-rrect
                resized_dst_img = resize_img_to_width(
                    resized_dst_img, self.recognizer.get_model_width()
                )
                h, w = resized_dst_img.shape[:2]
                merged_img[:h, :, :] = resized_dst_img

            s += w

            # pad background color
            merged_img[:, s : s + self.spacing, :] = self._get_padding_color(
                merged_img, s
            )
            s += self.spacing

            if sub_rrect_idx == len(line.sub_rrects) - 1:
                merged_img[:, s + 1 :, :] = 127
                merged_imgs.append(merged_img)
                merged_img = (
                    np.ones(
                        shape=(
                            self.recognizer.get_model_height(),
                            self.recognizer.get_model_width(),
                            3,
                        ),
                        dtype=np.uint8,
                    )
                    * 255
                )
                s = 0

        if self.debug:
            lines_dir = os.path.join(self.debug_dir, "detector/lines")
            if not os.path.exists(lines_dir):
                os.makedirs(lines_dir)

            cv2.imwrite(os.path.join(lines_dir, f"{line_idx}.jpg"), line_img)

            sub_rrect_dir = os.path.join(self.debug_dir, "detector/sub_rrects")
            if not os.path.exists(sub_rrect_dir):
                os.makedirs(sub_rrect_dir)

            for merged_img_idx, merged_img in enumerate(merged_imgs):
                image_path = os.path.join(
                    sub_rrect_dir, f"{line_idx}_{merged_img_idx}.jpg"
                )
                cv2.imwrite(image_path, merged_img)

        recogition_results = self.recognizer.infer_batch(merged_imgs)

        i = 0
        for sub_rrects, recognition_result in zip(
            merged_sub_rrect_list, recogition_results
        ):
            sub_text, sub_prob = recognition_result

            if sub_text == "" and not self.merge_boxes_inference:
                # should we check if this text is 45 degrees rotated
                # NOTE: only works when we do inference on each box independently
                sub_texts = []
                sub_probs = []
                for angle in [45, 135]:
                    rotated_img = rotate_image(merged_imgs[i], angle)
                    extra_result = self.recognizer.infer(rotated_img)
                    sub_texts.append(extra_result[0])
                    sub_probs.append(extra_result[1])
                sub_text = sub_texts[np.argmax(sub_probs)]

            if self.debug:
                print(f"Line {line_idx} - MergedImageIdx {i} - {sub_text} - {sub_prob}")

            if sub_text != "":
                if text != "":
                    text += " "
                text += sub_text
                prob *= sub_prob

            if not self.merge_boxes_inference:
                # sub_rrects should have only 1 object
                sub_rrects[0].text = sub_text

            i += 1

        return LanyOcrResult(text=text, prob=prob, line=line)

    def _get_padding_color(self, merged_img, s):
        # pick a color to set as background color between text-words
        margin = 2
        width = 2

        col1 = merged_img[margin : margin + width, :s, :]
        col2 = merged_img[
            self.recognizer.get_model_height()
            - margin
            - width : self.recognizer.get_model_height()
            - margin,
            :s,
            :,
        ]
        last_column = np.concatenate([col1, col2], axis=0)
        last_column = np.reshape(last_column, [-1, 3])

        b = last_column[:, 0]
        g = last_column[:, 1]
        r = last_column[:, 2]
        b = np.sort(b)
        g = np.sort(g)
        r = np.sort(r)
        b = b[len(b) // 2]
        g = g[len(g) // 2]
        r = r[len(r) // 2]

        return np.array([b, g, r], dtype=np.uint8)

    def _crop_and_rotate_rrect_horizontal(
        self, line_img, line_rrect, sub_lanyrrect: LanyOcrRRect, line_flipped: bool
    ):
        rrect_center = sub_lanyrrect.rrect[0]

        # rotate sub-rrect to horizontal line
        delta_x = abs(line_rrect[0][0] - rrect_center[0])
        delta_y = abs(line_rrect[0][1] - rrect_center[1])
        d = distance(line_rrect[0], rrect_center)

        # print(delta_x, delta_y, d)

        if 1.05 * delta_x > delta_y:
            if rrect_center[0] < line_rrect[0][0]:
                new_x = line_img.shape[1] // 2 - d
            else:
                new_x = line_img.shape[1] // 2 + d
        else:
            if rrect_center[1] < line_rrect[0][1]:
                new_x = line_img.shape[1] // 2 + d
            else:
                new_x = line_img.shape[1] // 2 - d

        if line_flipped:
            new_x = line_img.shape[1] - new_x

        new_y = line_img.shape[0] // 2
        w, h = sub_lanyrrect.rrect[1]

        new_x, new_y, w, h = int(new_x), int(new_y), int(w), int(h)

        if h > w:
            w, h = h, w

        left = max(0, new_x - w // 2)
        right = min(new_x + w // 2, line_img.shape[1] - 1)

        return line_img[:, left:right, :]

    def _is_line_flipped(self, line_img, thresh):
        # determine if a text is 180 degrees flipped or not
        flipped_count = 0
        num_tries = 3

        padded_imgs = []
        for padding_size in range(num_tries):
            padded_img = cv2.copyMakeBorder(
                line_img,
                padding_size,
                padding_size,
                padding_size,
                padding_size,
                cv2.BORDER_CONSTANT,
            )

            padded_imgs.append(padded_img)

        # run in batch
        for rotated_angle in self.angle_classifier.infer_batch(padded_imgs, thresh):
            if rotated_angle == 180:
                flipped_count += 1

        return flipped_count > num_tries // 2



from typing import List

from lanyocr.lanyocr_utils import LanyOcrRRect
from lanyocr.lanyocr_utils import LanyOcrTextLine
from lanyocr.text_merger import LanyOcrMerger


class LanyOcrNoMerger(LanyOcrMerger):
    def __init__(self) -> None:
        super().__init__()

    def merge_to_lines(
        self,
        rrects: List[LanyOcrRRect],
        merge_rotated: bool = True,
        merge_vertical: bool = True,
    ) -> List[LanyOcrTextLine]:
        lines: List[LanyOcrTextLine] = []

        for rrect in rrects:
            lines.append(
                LanyOcrTextLine(
                    sub_rrects=[rrect],
                    direction="",
                )
            )

        return lines



import os
from abc import ABC
from abc import abstractmethod
from typing import List
from typing import Tuple

from lanyocr.lanyocr_utils import LanyOcrRRect

MODULE_DICT = {
    "easyocr_craft": ["easyocr_craft.py", "EasyOcrCraft"],
    "paddleocr_en_ppocr_v3": ["paddleocr_en_ppocrv3.py", "PaddleOcrEnPPOCRV3"],
}


class LanyOcrDetector(ABC):
    def __init__(self, use_gpu: bool = True) -> None:
        self.use_gpu = use_gpu

    @abstractmethod
    def infer(self, image) -> List[LanyOcrRRect]:
        raise NotImplementedError


class LanyOcrDetectorFactory:
    @staticmethod
    def create(name: str, **kwargs) -> LanyOcrDetector:
        if name not in MODULE_DICT:
            raise ValueError("Invalid name")

        import importlib
        import inspect

        cur_dir = os.path.dirname(os.path.realpath(__file__))
        module_name = name
        module_path, class_name = MODULE_DICT[name]
        spec = importlib.util.spec_from_file_location(
            module_name, os.path.join(cur_dir, module_path)
        )
        _module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(_module)

        clsmembers = inspect.getmembers(_module, inspect.isclass)
        for clsmember in clsmembers:
            name, cls_def = clsmember
            if name == class_name:
                return cls_def(**kwargs)

        raise ValueError(f"Could not find class {class_name} in {module_path}")



import math
from typing import List

import cv2
import numpy as np
import onnxruntime
import pyclipper
from shapely.geometry import Polygon

from lanyocr.lanyocr_utils import LanyOcrRRect
from lanyocr.lanyocr_utils import download_model
from lanyocr.text_detector import LanyOcrDetector


class PaddleOcrEnPPOCRV3(LanyOcrDetector):
    def __init__(self, use_gpu: bool = True) -> None:
        super().__init__(use_gpu)

        print("Detector: PaddleOcrEnPPOCRV3")

        model_path = download_model("lanyocr-en-ppocrv3-det.onnx")

        # load model
        providers = ["CPUExecutionProvider"]
        if self.use_gpu:
            providers = ["CUDAExecutionProvider"]

        opts = onnxruntime.SessionOptions()
        opts.graph_optimization_level = (
            onnxruntime.GraphOptimizationLevel.ORT_ENABLE_EXTENDED
        )
        opts.log_severity_level = 3

        self.session = onnxruntime.InferenceSession(
            model_path, sess_options=opts, providers=providers
        )

        self.limit_side_len = 960
        self.limit_type = "max"
        self.scale = 1.0 / 255.0
        self.mean = [0.485, 0.456, 0.406]
        self.std = [0.229, 0.224, 0.225]

        self.thresh = 0.3
        self.max_candidates = 1000
        self.score_mode = "slow"
        self.box_thresh = 0.6
        self.min_size = 3
        self.unclip_ratio = 1.5

    def infer(self, image) -> List[LanyOcrRRect]:
        img, shape_list = self.preprocess(image)
        imgs = np.expand_dims(img, axis=0)
        shape_list = np.expand_dims(shape_list, axis=0)

        input_dict = {}
        input_dict["x"] = imgs

        preds = self.session.run(None, input_dict)[0]
        boxes = self.postprocess(preds, shape_list)[0]  # [num_boxes, 4, 2]

        rrects: List[LanyOcrRRect] = []
        for box in boxes:
            # make clock-wise order
            startidx = box.sum(axis=1).argmin()
            box = np.roll(box, 4 - startidx, 0)
            box = np.array(box)

            cnts = np.array(box).reshape((1, -1, 2))
            rrect = cv2.minAreaRect(cnts)

            rrects.append(
                LanyOcrRRect(
                    rrect=rrect,
                    points=np.reshape(box, [-1, 2]).tolist(),
                    direction="",
                )
            )

        return rrects

    # Internal functions
    def preprocess(self, img: np.ndarray):
        src_h, src_w, _ = img.shape

        if sum([src_h, src_w]) < 64:
            img = self.image_padding(img)

        img, [ratio_h, ratio_w] = self.resize_image_type0(img)

        # normalize
        img = (img.astype(np.float32) * self.scale - self.mean) / self.std

        img = np.transpose(img, (2, 0, 1))

        return img.astype(np.float32), np.array([src_h, src_w, ratio_h, ratio_w])

    def postprocess(self, pred, shape_list):
        pred = pred[:, 0, :, :]
        segmentation = pred > self.thresh
        segmentation = pred * segmentation.astype(np.float32)

        boxes_batch = []
        for batch_index in range(pred.shape[0]):
            src_h, src_w, ratio_h, ratio_w = shape_list[batch_index]
            mask = segmentation[batch_index]

            boxes, scores = self.boxes_from_bitmap(
                pred[batch_index], mask, src_w, src_h
            )

            boxes_batch.append(boxes)
        return boxes_batch

    def image_padding(self, im, value=0):
        h, w, c = im.shape
        im_pad = np.zeros((max(32, h), max(32, w), c), np.uint8) + value
        im_pad[:h, :w, :] = im
        return im_pad

    def resize_image_type0(self, img):
        """
        resize image to a size multiple of 32 which is required by the network
        args:
            img(array): array with shape [h, w, c]
        return(tuple):
            img, (ratio_h, ratio_w)
        """
        limit_side_len = self.limit_side_len
        h, w, c = img.shape

        # limit the max side
        if self.limit_type == "max":
            if max(h, w) > limit_side_len:
                if h > w:
                    ratio = float(limit_side_len) / h
                else:
                    ratio = float(limit_side_len) / w
            else:
                ratio = 1.0
        elif self.limit_type == "min":
            if min(h, w) < limit_side_len:
                if h < w:
                    ratio = float(limit_side_len) / h
                else:
                    ratio = float(limit_side_len) / w
            else:
                ratio = 1.0
        elif self.limit_type == "resize_long":
            ratio = float(limit_side_len) / max(h, w)
        else:
            raise Exception("not support limit type, image ")
        resize_h = int(h * ratio)
        resize_w = int(w * ratio)

        resize_h = max(int(round(resize_h / 32) * 32), 32)
        resize_w = max(int(round(resize_w / 32) * 32), 32)

        try:
            if int(resize_w) <= 0 or int(resize_h) <= 0:
                return None, (None, None)
            img = cv2.resize(img, (int(resize_w), int(resize_h)))
        except Exception as e:
            raise e

        ratio_h = resize_h / float(h)
        ratio_w = resize_w / float(w)
        return img, [ratio_h, ratio_w]

    def boxes_from_bitmap(self, pred, _bitmap, dest_width, dest_height):
        """
        _bitmap: single map with shape (1, H, W),
                whose values are binarized as {0, 1}
        """

        bitmap = _bitmap
        height, width = bitmap.shape

        outs = cv2.findContours(
            (bitmap * 255).astype(np.uint8), cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE
        )
        if len(outs) == 3:
            img, contours, _ = outs[0], outs[1], outs[2]
        elif len(outs) == 2:
            contours, _ = outs[0], outs[1]

        num_contours = min(len(contours), self.max_candidates)

        boxes = []
        scores = []
        for index in range(num_contours):
            contour = contours[index]
            points, sside = self.get_mini_boxes(contour)
            if sside < self.min_size:
                continue
            points = np.array(points)
            if self.score_mode == "fast":
                score = self.box_score_fast(pred, points.reshape(-1, 2))
            else:
                score = self.box_score_slow(pred, contour)
            if self.box_thresh > score:
                continue

            box = self.unclip(points, self.unclip_ratio).reshape(-1, 1, 2)
            box, sside = self.get_mini_boxes(box)
            if sside < self.min_size + 2:
                continue
            box = np.array(box)

            box[:, 0] = np.clip(np.round(box[:, 0] / width * dest_width), 0, dest_width)
            box[:, 1] = np.clip(
                np.round(box[:, 1] / height * dest_height), 0, dest_height
            )
            boxes.append(box.astype("int32"))
            scores.append(score)
        return np.array(boxes, dtype="int32"), scores

    def get_mini_boxes(self, contour):
        bounding_box = cv2.minAreaRect(contour)
        points = sorted(list(cv2.boxPoints(bounding_box)), key=lambda x: x[0])

        index_1, index_2, index_3, index_4 = 0, 1, 2, 3
        if points[1][1] > points[0][1]:
            index_1 = 0
            index_4 = 1
        else:
            index_1 = 1
            index_4 = 0
        if points[3][1] > points[2][1]:
            index_2 = 2
            index_3 = 3
        else:
            index_2 = 3
            index_3 = 2

        box = [points[index_1], points[index_2], points[index_3], points[index_4]]
        return box, min(bounding_box[1])

    def box_score_fast(self, bitmap, _box):
        """
        box_score_fast: use bbox mean score as the mean score
        """
        h, w = bitmap.shape[:2]
        box = _box.copy()
        xmin = np.clip(np.floor(box[:, 0].min()).astype("int32"), 0, w - 1)
        xmax = np.clip(np.ceil(box[:, 0].max()).astype("int32"), 0, w - 1)
        ymin = np.clip(np.floor(box[:, 1].min()).astype("int32"), 0, h - 1)
        ymax = np.clip(np.ceil(box[:, 1].max()).astype("int32"), 0, h - 1)

        mask = np.zeros((ymax - ymin + 1, xmax - xmin + 1), dtype=np.uint8)
        box[:, 0] = box[:, 0] - xmin
        box[:, 1] = box[:, 1] - ymin
        cv2.fillPoly(mask, box.reshape(1, -1, 2).astype("int32"), 1)
        return cv2.mean(bitmap[ymin : ymax + 1, xmin : xmax + 1], mask)[0]

    def box_score_slow(self, bitmap, contour):
        """
        box_score_slow: use polyon mean score as the mean score
        """
        h, w = bitmap.shape[:2]
        contour = contour.copy()
        contour = np.reshape(contour, (-1, 2))

        xmin = np.clip(np.min(contour[:, 0]), 0, w - 1)
        xmax = np.clip(np.max(contour[:, 0]), 0, w - 1)
        ymin = np.clip(np.min(contour[:, 1]), 0, h - 1)
        ymax = np.clip(np.max(contour[:, 1]), 0, h - 1)

        mask = np.zeros((ymax - ymin + 1, xmax - xmin + 1), dtype=np.uint8)

        contour[:, 0] = contour[:, 0] - xmin
        contour[:, 1] = contour[:, 1] - ymin

        cv2.fillPoly(mask, contour.reshape(1, -1, 2).astype("int32"), 1)
        return cv2.mean(bitmap[ymin : ymax + 1, xmin : xmax + 1], mask)[0]

    def unclip(self, box, unclip_ratio):
        poly = Polygon(box)
        distance = poly.area * unclip_ratio / poly.length
        offset = pyclipper.PyclipperOffset()
        offset.AddPath(box, pyclipper.JT_ROUND, pyclipper.ET_CLOSEDPOLYGON)
        expanded = np.array(offset.Execute(distance))
        return expanded



import math
import os
from enum import Enum
from typing import Any, List, Tuple

import cv2
import numpy as np
import requests
from pydantic import BaseModel


class LanyOcrMergeType(str, Enum):
    UPWARD = "upward"
    DOWNWARD = "downward"
    VERTICAL = "vertical"
    HORIZONTAL = "horizontal"


class LanyOcrBoundingBox(BaseModel):
    left: float
    top: float
    right: float
    bottom: float

    def area(self) -> float:
        return (self.right - self.left) * (self.bottom - self.top)


class LanyOcrRRect(BaseModel):
    rrect: Any
    points: List[List[float]]
    direction: str
    text: str = ""

    def getCenter(self) -> Tuple[float, float]:
        return (self.rrect[0][0], self.rrect[0][1])

    def getCenterX(self) -> float:
        return self.getCenter()[0]

    def getCenterY(self) -> float:
        return self.getCenter()[1]

    def getWidth(self) -> float:
        return self.rrect[1][0]

    def getHeight(self) -> float:
        return self.rrect[1][1]

    def getAngle(self) -> float:
        return self.rrect[2]

    def minX(self) -> float:
        return np.min([p[0] for p in self.points])

    def maxX(self) -> float:
        return np.max([p[0] for p in self.points])

    def minY(self) -> float:
        return np.min([p[1] for p in self.points])

    def maxY(self) -> float:
        return np.max([p[1] for p in self.points])

    def getBoundingBox(self) -> LanyOcrBoundingBox:
        box = cv2.boundingRect(
            np.expand_dims(np.array(self.points).astype(np.int32), 0)
        )
        return LanyOcrBoundingBox(
            left=box[0], top=box[1], right=box[0] + box[2], bottom=box[1] + box[3]
        )


class LanyOcrDirectionalBoxGroups(BaseModel):
    upward: List[LanyOcrRRect]
    downward: List[LanyOcrRRect]
    vertical: List[LanyOcrRRect]
    horizontal: List[LanyOcrRRect]


class LanyOcrTextLine(BaseModel):
    # rrect: LanyOcrRRect
    sub_rrects: List[LanyOcrRRect]
    direction: str = ""

    def avgAngle(self) -> float:
        angles = [rrect.rrect[2] for rrect in self.sub_rrects]
        return np.mean(angles)

    def avgWidth(self) -> float:
        widths = [rrect.rrect[1][0] for rrect in self.sub_rrects]
        return np.mean(widths)

    def avgHeight(self) -> float:
        heights = [rrect.rrect[1][1] for rrect in self.sub_rrects]
        return np.mean(heights)

    def get_rrect(self):
        points = []
        if (
            self.direction
            in ["", LanyOcrMergeType.HORIZONTAL, LanyOcrMergeType.VERTICAL]
            or len(self.sub_rrects) == 1
        ):
            for rrect in self.sub_rrects:
                points.extend(rrect.points)
        else:
            points.extend(self.sub_rrects[0].points)
            for rrect_idx in range(1, len(self.sub_rrects)):
                rrect = self.sub_rrects[rrect_idx]

                sum_xy = []
                for p in rrect.points:
                    sum_xy.append(p[0] + p[1])

                sorted_points = [x for _, x in sorted(zip(sum_xy, rrect.points))]
                points.append(sorted_points[-1])
                points.append(sorted_points[-2])
                points.append(sorted_points[-3])

        return cv2.minAreaRect(np.array([points]).astype(np.int32))


class LanyOcrResult(BaseModel):
    text: str
    prob: float
    line: LanyOcrTextLine


def iou(box1: LanyOcrBoundingBox, box2: LanyOcrBoundingBox) -> float:
    xx1 = max(box1.left, box2.left)
    yy1 = max(box1.top, box2.top)
    xx2 = min(box1.right, box2.right)
    yy2 = min(box1.bottom, box2.bottom)

    w = np.maximum(0.0, xx2 - xx1 + 1)
    h = np.maximum(0.0, yy2 - yy1 + 1)
    inter = w * h

    box1_area = box1.area()
    box2_area = box2.area()

    return inter / (box1_area + box2_area - inter)


def compute_angle(vector_1, vector_2):
    unit_vector_1 = vector_1 / np.linalg.norm(vector_1)
    unit_vector_2 = vector_2 / np.linalg.norm(vector_2)
    dot_product = np.dot(unit_vector_1, unit_vector_2)
    return np.arccos(dot_product) / np.pi * 180


def check_area_merge(rrect1: LanyOcrRRect, rrect2: LanyOcrRRect, thresh=0.75):
    points = []
    points.extend(rrect1.points)
    points.extend(rrect2.points)

    merged_rrect = cv2.minAreaRect(np.array([points]).astype(np.int32))

    rrect1_area = rrect1.getHeight() * rrect1.getWidth()
    rrect2_area = rrect2.getHeight() * rrect2.getWidth()
    merged_rrect_area = merged_rrect[1][0] * merged_rrect[1][1]

    ratio = (rrect1_area + rrect2_area) / merged_rrect_area
    return ratio > thresh


def check_merged_size(text_line: LanyOcrTextLine, rrect: LanyOcrRRect):
    text_line_rrect = text_line.get_rrect()
    text_line_points = cv2.boxPoints(text_line_rrect)

    points = []
    points.extend(text_line_points)
    points.extend(rrect.points)

    merged_rrect = cv2.minAreaRect(np.array([points]).astype(np.int32))
    w, h = merged_rrect[1]

    # merged rect should not have both h & w larger than each rrect
    if w > 1.5 * text_line_rrect[1][0] and h > 1.5 * text_line_rrect[1][1]:
        return False

    if w > 1.5 * rrect.getWidth() and h > 1.5 * rrect.getHeight():
        return False

    return True


def crop_rect(img, rrect):
    points = cv2.boxPoints(rrect)
    box = np.int0(np.array(points))

    min_x = np.min(box[:, 0])
    max_x = np.max(box[:, 0])
    min_y = np.min(box[:, 1])
    max_y = np.max(box[:, 1])

    cropped_img = img[min_y:max_y, min_x:max_x, :]
    return np.array(cropped_img)


def crop_rrect(img, rrect):
    if rrect[2] in [0, 90]:
        return crop_rect(img, rrect)

    padding_size = int(max(rrect[1][0], rrect[1][1])) // 2
    padded_img = cv2.copyMakeBorder(
        img,
        padding_size,
        padding_size,
        padding_size,
        padding_size,
        cv2.BORDER_CONSTANT,
    )

    m = cv2.getRotationMatrix2D(
        (rrect[0][0] + padding_size, rrect[0][1] + padding_size), rrect[2], 1
    )

    rotated_img = cv2.warpAffine(
        padded_img, m, (padded_img.shape[1], padded_img.shape[0])
    )

    rect = cv2.getRectSubPix(
        rotated_img,
        (int(rrect[1][0]), int(rrect[1][1])),
        (int(rrect[0][0] + padding_size), int(rrect[0][1] + padding_size)),
    )

    return rect


def rotate_image(image, angle):
    image_center = tuple(np.array(image.shape[1::-1]) / 2)
    rot_mat = cv2.getRotationMatrix2D(image_center, angle, 1.0)
    result = cv2.warpAffine(image, rot_mat, image.shape[1::-1], flags=cv2.INTER_LINEAR)
    return result


def rotate_image_at_center(image, center, angle):
    rot_mat = cv2.getRotationMatrix2D(center, angle, 1.0)
    result = cv2.warpAffine(image, rot_mat, image.shape[1::-1], flags=cv2.INTER_LINEAR)
    return result


def isascii(s):
    """Check if the characters in string s are in ASCII, U+0-U+7F."""
    return len(s.encode()) == len(s)


def resize_img_to_height(img, model_height=32):
    h = img.shape[0]
    w = img.shape[1]

    ratio = model_height / float(h)
    _w = int(ratio * w)
    return cv2.resize(img, (_w, model_height))


def resize_img_to_width(img, model_width=320):
    h = img.shape[0]
    w = img.shape[1]

    ratio = model_width / float(w)
    _h = int(ratio * h)
    return cv2.resize(img, (model_width, h))


def distance(p1, p2):
    return math.sqrt((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2)


def download_to_memory(url: str):
    import requests

    try:
        r = requests.get(url, allow_redirects=True)
        assert r.status_code == 200
        return r.content
    except Exception as e:
        raise e


def download_to_file(url: str, output_path: str):
    try:
        data = download_to_memory(url)
        open(output_path, "wb").write(data)
    except Exception as e:
        raise e


def download_model(model_file_name: str):
    from pathlib import Path

    LANYOCR_MODEL_DIR_PATH = os.path.join(Path.home(), ".LanyOCR")
    if not os.path.exists(LANYOCR_MODEL_DIR_PATH):
        os.makedirs(LANYOCR_MODEL_DIR_PATH)

    url = f"https://lanytek.com/models/{model_file_name}"
    output_path = os.path.join(LANYOCR_MODEL_DIR_PATH, model_file_name)

    if not os.path.exists(output_path):
        print(f"Downloading model to {output_path}")
        download_to_file(url, output_path)

    return output_path



from typing import List, Tuple
from lanyocr.text_merger import LanyOcrMerger
from lanyocr.lanyocr_utils import (
    LanyOcrMergeType,
    LanyOcrRRect,
    LanyOcrTextLine,
    check_area_merge,
    check_merged_size,
    iou,
)


class LanyOcrCraftBasedMerger(LanyOcrMerger):
    def __init__(self) -> None:
        super().__init__()

    def merge_to_lines(
        self,
        rrects: List[LanyOcrRRect],
        merge_rotated: bool = True,
        merge_vertical: bool = True,
    ) -> List[LanyOcrTextLine]:
        return self.merge_text_boxes(rrects, merge_rotated, merge_vertical)

    # Internal functions

    def check_downward_mergable(
        self,
        text_line: LanyOcrTextLine,
        rrect: LanyOcrRRect,
        width_thresh=0.5,
        iou_thresh=0.01,
    ) -> bool:
        last_rrect_in_line = text_line.sub_rrects[-1]

        if not check_area_merge(last_rrect_in_line, rrect, thresh=0.75):
            return False

        min_hw = min(
            min(last_rrect_in_line.getWidth(), last_rrect_in_line.getHeight()),
            min(rrect.getWidth(), rrect.getHeight()),
        )

        # we are going downward, cur box has to be on the right of prev box
        if rrect.getCenterX() - last_rrect_in_line.getCenterX() < 0.5 * min_hw:
            return False

        if rrect.getCenterY() - last_rrect_in_line.getCenterY() < 0.25 * min_hw:
            return False

        if (
            iou(rrect.getBoundingBox(), last_rrect_in_line.getBoundingBox())
            < iou_thresh
        ):
            return False

        max_last_x = last_rrect_in_line.maxX()
        min_cur_x = rrect.minX()

        if min_cur_x - max_last_x > min_hw * width_thresh:
            return False

        return True

    def check_upward_mergable(
        self,
        text_line: LanyOcrTextLine,
        rrect: LanyOcrRRect,
        height_thresh=0.75,
        width_thresh=0.5,
        angle_thresh=20,
        iou_thresh=0.005,
    ) -> bool:
        last_rrect_in_line = text_line.sub_rrects[-1]

        if not check_area_merge(last_rrect_in_line, rrect, thresh=0.75):
            return False

        min_hw = min(
            min(last_rrect_in_line.getWidth(), last_rrect_in_line.getHeight()),
            min(rrect.getWidth(), rrect.getHeight()),
        )

        # we are going upward, cur box has to be on the right of prev box
        if rrect.getCenterX() - last_rrect_in_line.getCenterX() < 0.5 * min_hw:
            return False

        if last_rrect_in_line.getCenterY() - rrect.getCenterY() < 0.25 * min_hw:
            return False

        last_rrect_bbox = last_rrect_in_line.getBoundingBox()
        last_rrect_bbox.left *= 0.975
        last_rrect_bbox.top *= 0.975
        last_rrect_bbox.right *= 1.025
        last_rrect_bbox.bottom *= 1.025

        cur_rrect_bbox = rrect.getBoundingBox()
        cur_rrect_bbox.left *= 0.975
        cur_rrect_bbox.top *= 0.975
        cur_rrect_bbox.right *= 1.025
        cur_rrect_bbox.bottom *= 1.025

        _iou = iou(last_rrect_bbox, cur_rrect_bbox)
        if _iou < iou_thresh:
            return False

        max_last_x = last_rrect_in_line.maxX()
        min_cur_x = rrect.minX()

        if min_cur_x - max_last_x > min_hw * width_thresh:
            return False

        return True

    def check_vertical_mergable(
        self,
        text_line: LanyOcrTextLine,
        rrect: LanyOcrRRect,
        iou_thresh=0.01,
    ) -> bool:
        last_rrect_in_line = text_line.sub_rrects[-1]

        if not check_area_merge(last_rrect_in_line, rrect, thresh=0.825):
            return False

        min_hw = min(
            min(last_rrect_in_line.getWidth(), last_rrect_in_line.getHeight()),
            min(rrect.getWidth(), rrect.getHeight()),
        )
        if abs(last_rrect_in_line.getCenterX() - rrect.getCenterX()) > 0.25 * min_hw:
            return False

        last_rrect_bbox = last_rrect_in_line.getBoundingBox()
        cur_rrect_bbox = rrect.getBoundingBox()
        # add some padding for both bboxes in y axis
        last_rrect_bbox.top *= 0.95
        last_rrect_bbox.bottom *= 1.05
        cur_rrect_bbox.top *= 0.95
        cur_rrect_bbox.bottom *= 1.05

        _iou = iou(last_rrect_bbox, cur_rrect_bbox)

        if _iou < iou_thresh:
            return False

        max_last_y = last_rrect_in_line.maxY()
        min_cur_y = rrect.minY()

        if abs(max_last_y - min_cur_y) > 0.5 * min_hw:
            return False

        if not check_merged_size(text_line, rrect):
            return False

        return True

    def check_horizontal_mergable(
        self,
        text_line: LanyOcrTextLine,
        rrect: LanyOcrRRect,
        iou_thresh=0.0025,
    ) -> bool:
        last_rrect_in_line = text_line.sub_rrects[-1]

        # print(last_rrect_in_line.rrect, rrect.rrect)

        if not check_area_merge(last_rrect_in_line, rrect, thresh=0.7):
            return False

        min_hw = min(
            min(last_rrect_in_line.getWidth(), last_rrect_in_line.getHeight()),
            min(rrect.getWidth(), rrect.getHeight()),
        )
        if abs(last_rrect_in_line.getCenterY() - rrect.getCenterY()) > 0.35 * min_hw:
            return False

        last_rrect_bbox = last_rrect_in_line.getBoundingBox()
        cur_rrect_bbox = rrect.getBoundingBox()

        # add some padding for both bboxes in y axis
        last_rrect_bbox.left *= 0.95
        last_rrect_bbox.right *= 1.05
        cur_rrect_bbox.left *= 0.95
        cur_rrect_bbox.right *= 1.05

        _iou = iou(last_rrect_bbox, cur_rrect_bbox)

        if _iou < iou_thresh:
            return False

        max_last_x = last_rrect_in_line.maxX()
        min_cur_x = rrect.minX()

        min_hw = min(text_line.avgHeight(), text_line.avgWidth())

        if abs(max_last_x - min_cur_x) > 0.5 * min_hw:
            return False

        return True

    def merge_text_boxes_step(
        self, rrects: List[LanyOcrRRect], merge_type: LanyOcrMergeType
    ) -> Tuple[List[LanyOcrTextLine], List[LanyOcrTextLine]]:

        multi_rrects_lines: List[LanyOcrTextLine] = []
        single_rrect_lines: List[LanyOcrTextLine] = []

        sorted_rrects: List[LanyOcrRRect] = []

        if merge_type in [LanyOcrMergeType.UPWARD]:
            sorted_rrects = sorted(
                rrects, reverse=True, key=lambda item: item.rrect[0][1]
            )
        elif merge_type in [LanyOcrMergeType.DOWNWARD]:
            sorted_rrects = sorted(
                rrects, reverse=False, key=lambda item: item.rrect[0][1]
            )
        elif merge_type in [LanyOcrMergeType.VERTICAL]:
            sorted_rrects = sorted(
                rrects, reverse=False, key=lambda item: item.rrect[0][1]
            )
        else:
            sorted_rrects = sorted(
                rrects, reverse=False, key=lambda item: item.rrect[0][0]
            )

        while sorted_rrects:
            base_rrect = sorted_rrects.pop(0)

            text_line = LanyOcrTextLine(sub_rrects=[base_rrect])
            text_line.direction = merge_type

            not_merged_rrects = []
            while sorted_rrects:
                cur_rrect = sorted_rrects.pop(0)
                mergable = False
                if merge_type == LanyOcrMergeType.DOWNWARD:
                    mergable = self.check_downward_mergable(text_line, cur_rrect)
                elif merge_type == LanyOcrMergeType.UPWARD:
                    mergable = self.check_upward_mergable(text_line, cur_rrect)
                elif merge_type == LanyOcrMergeType.VERTICAL:
                    mergable = self.check_vertical_mergable(text_line, cur_rrect)
                else:
                    mergable = self.check_horizontal_mergable(text_line, cur_rrect)

                if mergable:
                    text_line.sub_rrects.append(cur_rrect)
                else:
                    not_merged_rrects.append(cur_rrect)

            # sorted_rrects.extend(not_merged_rrects)
            sorted_rrects = not_merged_rrects

            if len(text_line.sub_rrects) > 1:
                multi_rrects_lines.append(text_line)
            else:
                single_rrect_lines.append(text_line)

        return (multi_rrects_lines, single_rrect_lines)

    def merge_text_boxes(
        self,
        rrects: List[LanyOcrRRect],
        merge_rotated: bool = True,
        merge_vertical: bool = False,
    ) -> List[LanyOcrTextLine]:
        results = []

        all_rrects = [rrect for rrect in rrects]

        (
            horizontal_multi_rrects_lines,
            horizontal_single_rrect_lines,
        ) = self.merge_text_boxes_step(all_rrects, LanyOcrMergeType.HORIZONTAL)
        results.extend(horizontal_multi_rrects_lines)

        if not merge_rotated and not merge_vertical:
            results.extend(horizontal_single_rrect_lines)
            return results

        all_rrects.clear()
        for line in horizontal_single_rrect_lines:
            all_rrects.extend(line.sub_rrects)

        if merge_rotated:
            (
                upward_multi_rrects_lines,
                upward_single_rrect_lines,
            ) = self.merge_text_boxes_step(all_rrects, LanyOcrMergeType.UPWARD)
            results.extend(upward_multi_rrects_lines)

            all_rrects.clear()
            for line in upward_single_rrect_lines:
                all_rrects.extend(line.sub_rrects)

            (
                downward_multi_rrects_lines,
                downward_single_rrect_lines,
            ) = self.merge_text_boxes_step(all_rrects, LanyOcrMergeType.DOWNWARD)
            results.extend(downward_multi_rrects_lines)

            if not merge_vertical:
                results.extend(downward_single_rrect_lines)
                return results

            all_rrects.clear()
            for line in downward_single_rrect_lines:
                all_rrects.extend(line.sub_rrects)

        (
            vertical_multi_rrects_lines,
            vertical_single_rrect_lines,
        ) = self.merge_text_boxes_step(all_rrects, LanyOcrMergeType.VERTICAL)

        results.extend(vertical_multi_rrects_lines)
        results.extend(vertical_single_rrect_lines)

        return results



from abc import ABC, abstractmethod
from enum import Enum
from typing import Dict

import cv2
import matplotlib.pyplot as plt
import numpy as np
from lanyocr import LanyOcr
from pydantic import BaseModel


class DatasetType(str, Enum):
    ICDAR2015 = "ICDAR2015"
    ICDAR2017 = "ICDAR2017"


class DetectorAccuracy(BaseModel):
    precision: float
    recall: float
    hmean: float


class RecognizerAccuracy(BaseModel):
    precision: float
    recall: float
    hmean: float


class OcrAccuracy(BaseModel):
    precision: float
    recall: float
    hmean: float


class LanyBenchmarker(ABC):
    def __init__(self, ocr: LanyOcr, dataset_path: str) -> None:
        self.dataset = []
        self.ocr = ocr
        self.dataset_path = dataset_path
        self.bounding = []
        self.predict = []
        self.chunk_size = 0

        # use ICDAR 2015 format for all kind of dataset
        self.id2imagePath: Dict[int, str] = {}
        self.id2gts: Dict[int, bytes] = {}
        self.id2preds: Dict[int, bytes] = {}

        self.load_dataset()

    @abstractmethod
    def load_dataset(self):
        """Load the dataset from the dataset_path into ICDAR 2015 format.
        Raises:
            NotImplementedError: _description_
        """

        raise NotImplementedError

    def compute_detector_accuracy(self) -> DetectorAccuracy:
        """Compute accuracy of only text detector
        Returns:
            DetectorAccuracy: _description_
        """

        """
            TODO: 
            - call self.ocr.detector.infer(...) on entire dataset to get bounding boxes of texts
            - compute precision/recall/fscore
            - return accuracy in the form of DetectorAccuracy
        """

        # run detector on every single image and store results as a string
        for id in self.id2imagePath:
            image_path = self.id2imagePath[id]
            image = cv2.imread(image_path)
            det_results = self.ocr.detector.infer(image)

            result_str = ""
            for det_result in det_results:
                points = det_result.points
                if len(points) != 4:
                    pass

                result_str += "{},{},".format(int(points[0][0]), int(points[0][1]))
                result_str += "{},{},".format(int(points[1][0]), int(points[1][1]))
                result_str += "{},{},".format(int(points[2][0]), int(points[2][1]))
                result_str += "{},{}".format(int(points[3][0]), int(points[3][1]))
                result_str += "\n"

            self.id2preds[id] = result_str.encode()

        from .detector_utils import evaluate_method

        results = evaluate_method(self.id2gts, self.id2preds)

        return DetectorAccuracy(
            precision=results["method"]["precision"],
            recall=results["method"]["recall"],
            hmean=results["method"]["hmean"],
        )

    def compute_recognizer_accuracy(self) -> RecognizerAccuracy:
        """Compute accuracy of only text recognizer
        Returns:
            RecognizerAccuracy: _description_
        """

        """ 
            TODO:
            - Assume the detector extracts all the box locations from dataset, extract all text images from dataset labels
            - run self.ocr.recognizer.infer(...) on all text images to predict text
            - return accuracy in the form of RecognizerAccuracy
        """

        _id2preds = {}

        for id in self.id2imagePath:
            image_path = self.id2imagePath[id]
            image = cv2.imread(image_path)

            result_str = ""
            gts = self.id2gts[id].decode().split("\n")
            for gt_idx, gt in enumerate(gts):
                if gt == "":
                    continue

                parts = gt.split(",")
                if parts[8] == "###":
                    # don't care
                    continue

                points = np.array([int(s) for s in parts[:8]]).reshape([1, 4, 2])
                rect = cv2.minAreaRect(points)
                # sub_img = crop_rect(image, rect)
                sub_img = self._crop_rect(image, rect)

                h, w = sub_img.shape[:2]

                if h == 0 or w == 0:
                    continue

                if h > 1.5 * w:
                    sub_img = cv2.rotate(sub_img, cv2.ROTATE_90_CLOCKWISE)

                # if self.ocr.angle_classifier.infer(sub_img) == 180:
                #     sub_img = cv2.rotate(sub_img, cv2.ROTATE_180)

                text_pred, _ = self.ocr.recognizer.infer(sub_img)

                # print(id, gt, text_pred)
                # if parts[8] != text_pred:
                #     print(id, gt, text_pred)
                #     # cv2.imwrite(f"./tmp/{id}_{gt_idx}.jpg", sub_img)

                for i in range(8):
                    result_str += parts[i] + ","
                result_str += text_pred + "\n"

            _id2preds[id] = result_str.encode()

        from .ocr_utils import evaluate_method

        results = evaluate_method(self.id2gts, _id2preds)

        return RecognizerAccuracy(
            precision=results["method"]["precision"],
            recall=results["method"]["recall"],
            hmean=results["method"]["hmean"],
        )

    def compute_e2e_accuracy(self) -> OcrAccuracy:
        """Compute the end-to-end accuracy of the ocr
        Returns:
            OcrAccuracy: _description_
        """

        """
            TODO:
            - run self.ocr.infer(...) on the entire dataset
            - return the accuracy in form of OcrAccuracy
        """

        # disable merge inference when running benchmark
        self.ocr.merge_boxes_inference = False

        for id in self.id2imagePath:
            image_path = self.id2imagePath[id]
            image = cv2.imread(image_path)

            ocr_lines = self.ocr.infer(image)

            result_str = ""
            for line in ocr_lines:
                for det_result in line.line.sub_rrects:
                    points = det_result.points
                    if len(points) != 4:
                        pass

                    result_str += "{},{},".format(int(points[0][0]), int(points[0][1]))
                    result_str += "{},{},".format(int(points[1][0]), int(points[1][1]))
                    result_str += "{},{},".format(int(points[2][0]), int(points[2][1]))
                    result_str += "{},{},".format(int(points[3][0]), int(points[3][1]))
                    result_str += det_result.text + "\n"

            self.id2preds[id] = result_str.encode()

        from .ocr_utils import evaluate_method

        results = evaluate_method(self.id2gts, self.id2preds)

        return OcrAccuracy(
            precision=results["method"]["precision"],
            recall=results["method"]["recall"],
            hmean=results["method"]["hmean"],
        )

    def _crop_rect(self, img, rect):
        # get the parameter of the small rectangle
        center = rect[0]
        size = rect[1]
        angle = rect[2]
        center, size = tuple(map(int, center)), tuple(map(int, size))

        # get row and col num in img
        rows, cols = img.shape[0], img.shape[1]

        M = cv2.getRotationMatrix2D(center, angle, 1)
        img_rot = cv2.warpAffine(img, M, (cols, rows))
        out = cv2.getRectSubPix(img_rot, size, center)

        return out



import math
from typing import List

import cv2
import numpy as np
import onnxruntime

from lanyocr.lanyocr_utils import LanyOcrRRect
from lanyocr.lanyocr_utils import download_model
from lanyocr.text_detector import LanyOcrDetector


class EasyOcrCraft(LanyOcrDetector):
    def __init__(self, use_gpu: bool = True) -> None:
        super().__init__(use_gpu)

        print("Detector: EasyOcrCraft")

        self.max_square_size = 1536

        model_path = download_model("lanyocr-easyocr-craft-detector.onnx")
        # load model
        providers = ["CPUExecutionProvider"]
        if self.use_gpu:
            providers = ["CUDAExecutionProvider"]

        opts = onnxruntime.SessionOptions()
        opts.graph_optimization_level = (
            onnxruntime.GraphOptimizationLevel.ORT_ENABLE_EXTENDED
        )
        opts.log_severity_level = 3

        self.session = onnxruntime.InferenceSession(
            model_path, sess_options=opts, providers=providers
        )

    def infer(self, image) -> List[LanyOcrRRect]:
        rrects: List[LanyOcrRRect] = []
        normed_image = self.preprocess(image)
        det_preds = self.session.run(["output"], {"input": [normed_image]})[0]
        det_polys = self.postprocess(det_preds[0], image)

        for poly in det_polys:
            cnts = np.array(poly).reshape((1, -1, 2))
            rrect = cv2.minAreaRect(cnts)

            rrects.append(
                LanyOcrRRect(
                    rrect=rrect,
                    points=np.reshape(poly, [-1, 2]).tolist(),
                    direction="",
                )
            )

        return rrects

    # Internal functions

    def load_image(self, image_path: str):
        image = cv2.imread(image_path)
        normed_image = self.preprocess(image)
        return normed_image, image

    def preprocess(self, image):
        img_resized, _, _ = self.resize_aspect_ratio(
            image, square_size=self.max_square_size, interpolation=cv2.INTER_LINEAR
        )
        img_resized = img_resized[:, :, ::-1]
        img_resized = self.normalizeMeanVariance(img_resized)
        img_resized = np.transpose(img_resized, [2, 0, 1])
        return img_resized

    def resize_aspect_ratio(self, img, square_size, interpolation, mag_ratio=1):
        height, width, channel = img.shape

        # magnify image size
        target_size = mag_ratio * max(height, width)

        # set original image size
        if target_size > square_size:
            target_size = square_size

        ratio = target_size / max(height, width)

        target_h, target_w = int(height * ratio), int(width * ratio)
        proc = cv2.resize(img, (target_w, target_h), interpolation=interpolation)

        # make canvas and paste image
        target_h32, target_w32 = target_h, target_w
        if target_h % 32 != 0:
            target_h32 = target_h + (32 - target_h % 32)
        if target_w % 32 != 0:
            target_w32 = target_w + (32 - target_w % 32)
        resized = np.zeros((target_h32, target_w32, channel), dtype=np.float32)
        resized[0:target_h, 0:target_w, :] = proc
        target_h, target_w = target_h32, target_w32

        size_heatmap = (int(target_w / 2), int(target_h / 2))

        return resized, ratio, size_heatmap

    def normalizeMeanVariance(
        self, in_img, mean=(0.485, 0.456, 0.406), variance=(0.229, 0.224, 0.225)
    ):
        # should be RGB order
        img = in_img.copy().astype(np.float32)

        img -= np.array(
            [mean[0] * 255.0, mean[1] * 255.0, mean[2] * 255.0], dtype=np.float32
        )
        img /= np.array(
            [variance[0] * 255.0, variance[1] * 255.0, variance[2] * 255.0],
            dtype=np.float32,
        )
        return img

    # ref: https://github.com/JaidedAI/EasyOCR/blob/master/easyocr/detection.py
    def postprocess(
        self,
        preds,
        original_img,
        max_img_size: int = 1536,
        text_threshold: float = 0.7,
        low_text_threshold: float = 0.33,
        # configs to detect more texts
        # text_threshold: float = 0.6,
        # low_text_threshold: float = 0.25,
        link_text_threshold: float = 0.125,
    ):
        height, width = original_img.shape[:2]

        target_size = max(height, width)
        if target_size > max_img_size:
            target_size = max_img_size

        target_ratio = float(target_size) / max(height, width)
        ratio_h = ratio_w = 1.0 / target_ratio

        score_text = preds[:, :, 0]
        score_link = preds[:, :, 1]

        # Post-processing
        estimate_num_chars = False
        boxes, polys, mapper = self.getDetBoxes(
            score_text,
            score_link,
            text_threshold=text_threshold,
            link_threshold=link_text_threshold,
            low_text=low_text_threshold,
            poly=False,
            estimate_num_chars=estimate_num_chars,
        )

        # coordinate adjustment
        boxes = self.adjustResultCoordinates(boxes, ratio_w, ratio_h)
        polys = self.adjustResultCoordinates(polys, ratio_w, ratio_h)

        for k in range(len(polys)):
            if polys[k] is None:
                polys[k] = boxes[k]

        result = []
        for i, box in enumerate(polys):
            poly = np.array(box).astype(np.int32).reshape((-1))
            result.append(poly)

        return result

    # ref: https://github.com/JaidedAI/EasyOCR/blob/master/easyocr/craft_utils.py
    def getDetBoxes(
        self,
        textmap,
        linkmap,
        text_threshold,
        link_threshold,
        low_text,
        poly=False,
        estimate_num_chars=False,
    ):
        if poly and estimate_num_chars:
            raise Exception(
                "Estimating the number of characters not currently supported with poly."
            )
        boxes, labels, mapper = self.getDetBoxes_core(
            textmap,
            linkmap,
            text_threshold,
            link_threshold,
            low_text,
            estimate_num_chars,
        )

        if poly:
            # polys = getPoly_core(boxes, labels, mapper, linkmap)
            pass
        else:
            polys = [None] * len(boxes)

        return boxes, polys, mapper

    # ref: https://github.com/JaidedAI/EasyOCR/blob/master/easyocr/craft_utils.py

    def getDetBoxes_core(
        self,
        textmap,
        linkmap,
        text_threshold,
        link_threshold,
        low_text,
        estimate_num_chars=False,
    ):
        # prepare data
        linkmap = linkmap.copy()
        textmap = textmap.copy()
        img_h, img_w = textmap.shape

        """ labeling method """
        ret, text_score = cv2.threshold(textmap, low_text, 1, 0)
        ret, link_score = cv2.threshold(linkmap, link_threshold, 1, 0)

        text_score_comb = np.clip(text_score + link_score, 0, 1)
        nLabels, labels, stats, centroids = cv2.connectedComponentsWithStats(
            text_score_comb.astype(np.uint8), connectivity=4
        )

        det = []
        mapper = []
        for k in range(1, nLabels):
            # size filtering
            size = stats[k, cv2.CC_STAT_AREA]
            if size < 10:
                continue

            # thresholding
            if np.max(textmap[labels == k]) < text_threshold:
                continue

            # make segmentation map
            segmap = np.zeros(textmap.shape, dtype=np.uint8)
            segmap[labels == k] = 255
            mapper.append(k)
            segmap[
                np.logical_and(link_score == 1, text_score == 0)
            ] = 0  # remove link area
            x, y = stats[k, cv2.CC_STAT_LEFT], stats[k, cv2.CC_STAT_TOP]
            w, h = stats[k, cv2.CC_STAT_WIDTH], stats[k, cv2.CC_STAT_HEIGHT]
            niter = int(math.sqrt(size * min(w, h) / (w * h)) * 2)
            sx, ex, sy, ey = x - niter, x + w + niter + 1, y - niter, y + h + niter + 1
            # boundary check
            if sx < 0:
                sx = 0
            if sy < 0:
                sy = 0
            if ex >= img_w:
                ex = img_w
            if ey >= img_h:
                ey = img_h
            kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (1 + niter, 1 + niter))
            segmap[sy:ey, sx:ex] = cv2.dilate(segmap[sy:ey, sx:ex], kernel)

            # make box
            np_contours = (
                np.roll(np.array(np.where(segmap != 0)), 1, axis=0)
                .transpose()
                .reshape(-1, 2)
            )
            rectangle = cv2.minAreaRect(np_contours)
            box = cv2.boxPoints(rectangle)

            # align diamond-shape
            w, h = np.linalg.norm(box[0] - box[1]), np.linalg.norm(box[1] - box[2])
            box_ratio = max(w, h) / (min(w, h) + 1e-5)
            if abs(1 - box_ratio) <= 0.1:
                l, r = min(np_contours[:, 0]), max(np_contours[:, 0])
                t, b = min(np_contours[:, 1]), max(np_contours[:, 1])
                box = np.array([[l, t], [r, t], [r, b], [l, b]], dtype=np.float32)

            # make clock-wise order
            startidx = box.sum(axis=1).argmin()
            box = np.roll(box, 4 - startidx, 0)
            box = np.array(box)

            det.append(box)

        return det, labels, mapper

    # ref: https://github.com/JaidedAI/EasyOCR/blob/master/easyocr/craft_utils.py
    def adjustResultCoordinates(self, polys, ratio_w, ratio_h, ratio_net=2):
        if len(polys) > 0:
            polys = np.array(polys)
            for k in range(len(polys)):
                if polys[k] is not None:
                    polys[k] *= (ratio_w * ratio_net, ratio_h * ratio_net)
        return polys



import json
import math
import os
from typing import List, Tuple

import cv2
import numpy as np
import onnxruntime
from lanyocr.text_recognizer import LanyOcrRecognizer


class PaddleOcrBase(LanyOcrRecognizer):
    def __init__(
        self,
        use_gpu: bool = True,
        model_ignored_tokens: List[int] = [],
        model_characters: List[str] = [],
        accepted_characters: List[str] = [],
        model_path: str = "",
        model_w: int = 320,
        model_h: int = 32,
    ) -> None:
        super().__init__(use_gpu)

        self.model_w = model_w
        self.model_h = model_h
        self.model_ignored_tokens = model_ignored_tokens
        self.model_characters = model_characters
        self.accepted_characters = accepted_characters
        self.model_path = model_path

        # load model
        providers = ["CPUExecutionProvider"]
        if self.use_gpu:
            providers = ["CUDAExecutionProvider"]

        opts = onnxruntime.SessionOptions()
        opts.graph_optimization_level = (
            onnxruntime.GraphOptimizationLevel.ORT_ENABLE_EXTENDED
        )

        self.session = onnxruntime.InferenceSession(
            self.model_path, sess_options=opts, providers=providers
        )

    def infer(self, bgr_img) -> Tuple[str, float]:
        norm_img = self.normalize_img(bgr_img)

        preds = self.session.run(None, {"x": [norm_img]})[0][0]

        preds_idx = preds.argmax(axis=1)
        preds_prob = preds.max(axis=1)
        results = self.decode([preds_idx], [preds_prob], True)
        text, prob = results[0]
        return text, prob

    def infer_batch(self, bgr_images) -> List[Tuple[str, float]]:
        batch_results = []
        normed_imgs = [self.normalize_img(img) for img in bgr_images]

        for idx in range(0, len(normed_imgs), self.max_batch_size):
            batch_preds = self.session.run(
                None, {"x": normed_imgs[idx : idx + self.max_batch_size]}
            )[0]

            for preds in batch_preds:
                preds_idx = preds.argmax(axis=1)
                preds_prob = preds.max(axis=1)
                results = self.decode([preds_idx], [preds_prob], True)
                batch_results.append(results[0])

        return batch_results

    def get_model_height(self) -> int:
        return self.model_h

    def get_model_width(self) -> int:
        return self.model_w

    def normalize_img(self, bgr_img):
        h, w = bgr_img.shape[:2]

        if h != self.model_h or w != self.model_w:
            ratio = w / float(h)

            if math.ceil(self.model_h * ratio) > self.model_w:
                resized_w = self.model_w
            else:
                resized_w = int(math.ceil(self.model_h * ratio))

            resized_image = cv2.resize(bgr_img, (resized_w, self.model_h))
        else:
            resized_image = bgr_img

        resized_image = resized_image.astype("float32")
        resized_image = resized_image.transpose((2, 0, 1)) / 255.0
        resized_image -= 0.5
        resized_image /= 0.5

        if h != self.model_h or w != self.model_w:
            final_img = np.zeros(
                shape=(3, self.model_h, self.model_w), dtype=np.float32
            )
            final_img[:, :, :resized_w] = resized_image
        else:
            final_img = resized_image

        return final_img

    def decode(self, text_index, text_prob=None, is_remove_duplicate=False):
        """convert text-index into text-label."""
        result_list = []
        batch_size = len(text_index)
        for batch_idx in range(batch_size):
            char_list = []
            conf_list = []
            for idx in range(len(text_index[batch_idx])):
                if text_index[batch_idx][idx] in self.model_ignored_tokens:
                    continue

                if is_remove_duplicate:
                    # only for predict
                    if (
                        idx > 0
                        and text_index[batch_idx][idx - 1] == text_index[batch_idx][idx]
                    ):
                        continue

                detected_char = self.model_characters[int(text_index[batch_idx][idx])]

                # filter out some unused characters
                if (
                    self.accepted_characters
                    and detected_char not in self.accepted_characters
                ):
                    continue

                char_list.append(detected_char)
                if text_prob is not None:
                    conf_list.append(text_prob[batch_idx][idx])
                else:
                    conf_list.append(1)

            text = "".join(char_list)
            prob = 1.0 if conf_list else 0
            for p in conf_list:
                prob *= p

            result_list.append((text, prob))
        return result_list



import re
from collections import namedtuple

import numpy as np
import Polygon as plg


def decode_utf8(raw):
    """
    Returns a Unicode object on success, or None on failure
    """
    try:
        return raw.decode("utf-8-sig", errors="replace")
    except:
        return None


def validate_clockwise_points(points):
    """
    Validates that the points are in clockwise order.
    """
    edge = []
    for i in range(len(points) // 2):
        edge.append(
            (int(points[(i + 1) * 2 % len(points)]) - int(points[i * 2]))
            * (int(points[((i + 1) * 2 + 1) % len(points)]) + int(points[i * 2 + 1]))
        )
    if sum(edge) > 0:
        raise Exception(
            "Points are not clockwise. The coordinates of bounding points have to be given in clockwise order. Regarding the correct interpretation of 'clockwise' remember that the image coordinate system used is the standard one, with the image origin at the upper left, the X axis extending to the right and Y axis extending downwards."
        )


def validate_point_inside_bounds(x, y, imWidth, imHeight):
    if x < 0 or x > imWidth:
        raise Exception(
            "X value (%s) not valid. Image dimensions: (%s,%s)" % (x, imWidth, imHeight)
        )
    if y < 0 or y > imHeight:
        raise Exception(
            "Y value (%s)  not valid. Image dimensions: (%s,%s) Sample: %s Line:%s"
            % (y, imWidth, imHeight)
        )


def get_tl_line_values(
    line,
    LTRB=True,
    withTranscription=False,
    withConfidence=False,
    imWidth=0,
    imHeight=0,
):
    """
    Validate the format of the line. If the line is not valid an exception will be raised.
    If maxWidth and maxHeight are specified, all points must be inside the imgage bounds.
    Posible values are:
    LTRB=True: xmin,ymin,xmax,ymax[,confidence][,transcription]
    LTRB=False: x1,y1,x2,y2,x3,y3,x4,y4[,confidence][,transcription]
    Returns values from a textline. Points , [Confidences], [Transcriptions]
    """
    confidence = 0.0
    transcription = ""
    points = []

    numPoints = 4

    if LTRB:

        numPoints = 4

        if withTranscription and withConfidence:
            m = re.match(
                r"^\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*([0-9]+)\s*,\s*([0-9]+)\s*,\s*([0-1].?[0-9]*)\s*,(.*)$",
                line,
            )
            if m == None:
                m = re.match(
                    r"^\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*([0-9]+)\s*,\s*([0-9]+)\s*,\s*([0-1].?[0-9]*)\s*,(.*)$",
                    line,
                )
                raise Exception(
                    "Format incorrect. Should be: xmin,ymin,xmax,ymax,confidence,transcription"
                )
        elif withConfidence:
            m = re.match(
                r"^\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*([0-9]+)\s*,\s*([0-9]+)\s*,\s*([0-1].?[0-9]*)\s*$",
                line,
            )
            if m == None:
                raise Exception(
                    "Format incorrect. Should be: xmin,ymin,xmax,ymax,confidence"
                )
        elif withTranscription:
            m = re.match(
                r"^\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*([0-9]+)\s*,\s*([0-9]+)\s*,(.*)$",
                line,
            )
            if m == None:
                raise Exception(
                    "Format incorrect. Should be: xmin,ymin,xmax,ymax,transcription"
                )
        else:
            m = re.match(
                r"^\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*([0-9]+)\s*,\s*([0-9]+)\s*,?\s*$",
                line,
            )
            if m == None:
                raise Exception("Format incorrect. Should be: xmin,ymin,xmax,ymax")

        xmin = int(m.group(1))
        ymin = int(m.group(2))
        xmax = int(m.group(3))
        ymax = int(m.group(4))
        if xmax < xmin:
            raise Exception("Xmax value (%s) not valid (Xmax < Xmin)." % (xmax))
        if ymax < ymin:
            raise Exception("Ymax value (%s)  not valid (Ymax < Ymin)." % (ymax))

        points = [float(m.group(i)) for i in range(1, (numPoints + 1))]

        if imWidth > 0 and imHeight > 0:
            validate_point_inside_bounds(xmin, ymin, imWidth, imHeight)
            validate_point_inside_bounds(xmax, ymax, imWidth, imHeight)

    else:
        numPoints = 8

        if withTranscription and withConfidence:
            m = re.match(
                r"^\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*([0-1].?[0-9]*)\s*,(.*)$",
                line,
            )
            if m == None:
                raise Exception(
                    "Format incorrect. Should be: x1,y1,x2,y2,x3,y3,x4,y4,confidence,transcription"
                )
        elif withConfidence:
            m = re.match(
                r"^\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*([0-1].?[0-9]*)\s*$",
                line,
            )
            if m == None:
                raise Exception(
                    "Format incorrect. Should be: x1,y1,x2,y2,x3,y3,x4,y4,confidence"
                )
        elif withTranscription:
            m = re.match(
                r"^\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,(.*)$",
                line,
            )
            if m == None:
                raise Exception(
                    "Format incorrect. Should be: x1,y1,x2,y2,x3,y3,x4,y4,transcription"
                )
        else:
            m = re.match(
                r"^\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*,\s*(-?[0-9]+)\s*$",
                line,
            )
            if m == None:
                raise Exception("Format incorrect. Should be: x1,y1,x2,y2,x3,y3,x4,y4")

        points = [float(m.group(i)) for i in range(1, (numPoints + 1))]

        validate_clockwise_points(points)

        if imWidth > 0 and imHeight > 0:
            validate_point_inside_bounds(points[0], points[1], imWidth, imHeight)
            validate_point_inside_bounds(points[2], points[3], imWidth, imHeight)
            validate_point_inside_bounds(points[4], points[5], imWidth, imHeight)
            validate_point_inside_bounds(points[6], points[7], imWidth, imHeight)

    if withConfidence:
        try:
            confidence = float(m.group(numPoints + 1))
        except ValueError:
            raise Exception("Confidence value must be a float")

    if withTranscription:
        posTranscription = numPoints + (2 if withConfidence else 1)
        transcription = m.group(posTranscription)
        m2 = re.match(r"^\s*\"(.*)\"\s*$", transcription)
        if (
            m2 != None
        ):  # Transcription with double quotes, we extract the value and replace escaped characters
            transcription = m2.group(1).replace("\\\\", "\\").replace('\\"', '"')

    return points, confidence, transcription


def get_tl_line_values_from_file_contents(
    content,
    CRLF=True,
    LTRB=True,
    withTranscription=False,
    withConfidence=False,
    imWidth=0,
    imHeight=0,
    sort_by_confidences=True,
):
    """
    Returns all points, confindences and transcriptions of a file in lists. Valid line formats:
    xmin,ymin,xmax,ymax,[confidence],[transcription]
    x1,y1,x2,y2,x3,y3,x4,y4,[confidence],[transcription]
    """
    pointsList = []
    transcriptionsList = []
    confidencesList = []

    lines = content.split("\r\n" if CRLF else "\n")
    for line in lines:
        line = line.replace("\r", "").replace("\n", "")
        if line != "":
            points, confidence, transcription = get_tl_line_values(
                line, LTRB, withTranscription, withConfidence, imWidth, imHeight
            )
            pointsList.append(points)
            transcriptionsList.append(transcription)
            confidencesList.append(confidence)

    if withConfidence and len(confidencesList) > 0 and sort_by_confidences:
        import numpy as np

        sorted_ind = np.argsort(-np.array(confidencesList))
        confidencesList = [confidencesList[i] for i in sorted_ind]
        pointsList = [pointsList[i] for i in sorted_ind]
        transcriptionsList = [transcriptionsList[i] for i in sorted_ind]

    return pointsList, confidencesList, transcriptionsList


def evaluate_method(
    gt,
    subm,
    evaluationParams={
        "IOU_CONSTRAINT": 0.5,
        "AREA_PRECISION_CONSTRAINT": 0.5,
        "GT_SAMPLE_NAME_2_ID": "gt_img_([0-9]+).txt",
        "DET_SAMPLE_NAME_2_ID": "res_img_([0-9]+).txt",
        "LTRB": False,
        "CRLF": False,
        "CONFIDENCES": False,
        "PER_SAMPLE_RESULTS": True,
    },
):
    """
    Method evaluate_method: evaluate method and returns the results
        Results. Dictionary with the following values:
        - method (required)  Global method metrics. Ex: { 'Precision':0.8,'Recall':0.9 }
        - samples (optional) Per sample metrics. Ex: {'sample1' : { 'Precision':0.8,'Recall':0.9 } , 'sample2' : { 'Precision':0.8,'Recall':0.9 }
    """

    def polygon_from_points(points):
        """
        Returns a Polygon object to use with the Polygon2 class from a list of 8 points: x1,y1,x2,y2,x3,y3,x4,y4
        """
        resBoxes = np.empty([1, 8], dtype="int32")
        resBoxes[0, 0] = int(points[0])
        resBoxes[0, 4] = int(points[1])
        resBoxes[0, 1] = int(points[2])
        resBoxes[0, 5] = int(points[3])
        resBoxes[0, 2] = int(points[4])
        resBoxes[0, 6] = int(points[5])
        resBoxes[0, 3] = int(points[6])
        resBoxes[0, 7] = int(points[7])
        pointMat = resBoxes[0].reshape([2, 4]).T
        return plg.Polygon(pointMat)

    def rectangle_to_polygon(rect):
        resBoxes = np.empty([1, 8], dtype="int32")
        resBoxes[0, 0] = int(rect.xmin)
        resBoxes[0, 4] = int(rect.ymax)
        resBoxes[0, 1] = int(rect.xmin)
        resBoxes[0, 5] = int(rect.ymin)
        resBoxes[0, 2] = int(rect.xmax)
        resBoxes[0, 6] = int(rect.ymin)
        resBoxes[0, 3] = int(rect.xmax)
        resBoxes[0, 7] = int(rect.ymax)

        pointMat = resBoxes[0].reshape([2, 4]).T

        return plg.Polygon(pointMat)

    def rectangle_to_points(rect):
        points = [
            int(rect.xmin),
            int(rect.ymax),
            int(rect.xmax),
            int(rect.ymax),
            int(rect.xmax),
            int(rect.ymin),
            int(rect.xmin),
            int(rect.ymin),
        ]
        return points

    def get_union(pD, pG):
        areaA = pD.area()
        areaB = pG.area()
        return areaA + areaB - get_intersection(pD, pG)

    def get_intersection_over_union(pD, pG):
        try:
            return get_intersection(pD, pG) / get_union(pD, pG)
        except:
            return 0

    def get_intersection(pD, pG):
        pInt = pD & pG
        if len(pInt) == 0:
            return 0
        return pInt.area()

    def compute_ap(confList, matchList, numGtCare):
        correct = 0
        AP = 0
        if len(confList) > 0:
            confList = np.array(confList)
            matchList = np.array(matchList)
            sorted_ind = np.argsort(-confList)
            confList = confList[sorted_ind]
            matchList = matchList[sorted_ind]
            for n in range(len(confList)):
                match = matchList[n]
                if match:
                    correct += 1
                    AP += float(correct) / (n + 1)

            if numGtCare > 0:
                AP /= numGtCare

        return AP

    perSampleMetrics = {}

    matchedSum = 0

    Rectangle = namedtuple("Rectangle", "xmin ymin xmax ymax")

    # print(list(gt.keys())[0], gt[list(gt.keys())[0]])
    # print(list(subm.keys())[0], subm[list(subm.keys())[0]])
    # print(type(subm))

    numGlobalCareGt = 0
    numGlobalCareDet = 0

    arrGlobalConfidences = []
    arrGlobalMatches = []

    for resFile in gt:

        gtFile = decode_utf8(gt[resFile])
        recall = 0
        precision = 0
        hmean = 0

        detMatched = 0

        iouMat = np.empty([1, 1])

        gtPols = []
        detPols = []

        gtPolPoints = []
        detPolPoints = []

        # Array of Ground Truth Polygons' keys marked as don't Care
        gtDontCarePolsNum = []
        # Array of Detected Polygons' matched with a don't Care GT
        detDontCarePolsNum = []

        pairs = []
        detMatchedNums = []

        arrSampleConfidences = []
        arrSampleMatch = []
        sampleAP = 0

        evaluationLog = ""

        (pointsList, _, transcriptionsList,) = get_tl_line_values_from_file_contents(
            gtFile, evaluationParams["CRLF"], evaluationParams["LTRB"], True, False
        )
        for n in range(len(pointsList)):
            points = pointsList[n]
            transcription = transcriptionsList[n]
            dontCare = transcription == "###"
            if evaluationParams["LTRB"]:
                gtRect = Rectangle(*points)
                gtPol = rectangle_to_polygon(gtRect)
            else:
                gtPol = polygon_from_points(points)
            gtPols.append(gtPol)
            gtPolPoints.append(points)
            if dontCare:
                gtDontCarePolsNum.append(len(gtPols) - 1)

        evaluationLog += (
            "GT polygons: "
            + str(len(gtPols))
            + (
                " (" + str(len(gtDontCarePolsNum)) + " don't care)\n"
                if len(gtDontCarePolsNum) > 0
                else "\n"
            )
        )

        if resFile in subm:

            detFile = decode_utf8(subm[resFile])

            (pointsList, confidencesList, _,) = get_tl_line_values_from_file_contents(
                detFile,
                evaluationParams["CRLF"],
                evaluationParams["LTRB"],
                False,
                evaluationParams["CONFIDENCES"],
            )
            for n in range(len(pointsList)):
                points = pointsList[n]

                if evaluationParams["LTRB"]:
                    detRect = Rectangle(*points)
                    detPol = rectangle_to_polygon(detRect)
                else:
                    detPol = polygon_from_points(points)
                detPols.append(detPol)
                detPolPoints.append(points)
                if len(gtDontCarePolsNum) > 0:
                    for dontCarePol in gtDontCarePolsNum:
                        dontCarePol = gtPols[dontCarePol]
                        intersected_area = get_intersection(dontCarePol, detPol)
                        pdDimensions = detPol.area()
                        precision = (
                            0 if pdDimensions == 0 else intersected_area / pdDimensions
                        )
                        if precision > evaluationParams["AREA_PRECISION_CONSTRAINT"]:
                            detDontCarePolsNum.append(len(detPols) - 1)
                            break

            evaluationLog += (
                "DET polygons: "
                + str(len(detPols))
                + (
                    " (" + str(len(detDontCarePolsNum)) + " don't care)\n"
                    if len(detDontCarePolsNum) > 0
                    else "\n"
                )
            )

            if len(gtPols) > 0 and len(detPols) > 0:
                # Calculate IoU and precision matrixs
                outputShape = [len(gtPols), len(detPols)]
                iouMat = np.empty(outputShape)
                gtRectMat = np.zeros(len(gtPols), np.int8)
                detRectMat = np.zeros(len(detPols), np.int8)
                for gtNum in range(len(gtPols)):
                    for detNum in range(len(detPols)):
                        pG = gtPols[gtNum]
                        pD = detPols[detNum]
                        iouMat[gtNum, detNum] = get_intersection_over_union(pD, pG)

                for gtNum in range(len(gtPols)):
                    for detNum in range(len(detPols)):
                        if (
                            gtRectMat[gtNum] == 0
                            and detRectMat[detNum] == 0
                            and gtNum not in gtDontCarePolsNum
                            and detNum not in detDontCarePolsNum
                        ):
                            if (
                                iouMat[gtNum, detNum]
                                > evaluationParams["IOU_CONSTRAINT"]
                            ):
                                gtRectMat[gtNum] = 1
                                detRectMat[detNum] = 1
                                detMatched += 1
                                pairs.append({"gt": gtNum, "det": detNum})
                                detMatchedNums.append(detNum)
                                evaluationLog += (
                                    "Match GT #"
                                    + str(gtNum)
                                    + " with Det #"
                                    + str(detNum)
                                    + "\n"
                                )

            if evaluationParams["CONFIDENCES"]:
                for detNum in range(len(detPols)):
                    if detNum not in detDontCarePolsNum:
                        # we exclude the don't care detections
                        match = detNum in detMatchedNums

                        arrSampleConfidences.append(confidencesList[detNum])
                        arrSampleMatch.append(match)

                        arrGlobalConfidences.append(confidencesList[detNum])
                        arrGlobalMatches.append(match)

        numGtCare = len(gtPols) - len(gtDontCarePolsNum)
        numDetCare = len(detPols) - len(detDontCarePolsNum)
        if numGtCare == 0:
            recall = float(1)
            precision = float(0) if numDetCare > 0 else float(1)
            sampleAP = precision
        else:
            recall = float(detMatched) / numGtCare
            precision = 0 if numDetCare == 0 else float(detMatched) / numDetCare
            if (
                evaluationParams["CONFIDENCES"]
                and evaluationParams["PER_SAMPLE_RESULTS"]
            ):
                sampleAP = compute_ap(arrSampleConfidences, arrSampleMatch, numGtCare)

        hmean = (
            0
            if (precision + recall) == 0
            else 2.0 * precision * recall / (precision + recall)
        )

        matchedSum += detMatched
        numGlobalCareGt += numGtCare
        numGlobalCareDet += numDetCare

        if evaluationParams["PER_SAMPLE_RESULTS"]:
            perSampleMetrics[resFile] = {
                "precision": precision,
                "recall": recall,
                "hmean": hmean,
                "pairs": pairs,
                "AP": sampleAP,
                "iouMat": [] if len(detPols) > 100 else iouMat.tolist(),
                "gtPolPoints": gtPolPoints,
                "detPolPoints": detPolPoints,
                "gtDontCare": gtDontCarePolsNum,
                "detDontCare": detDontCarePolsNum,
                "evaluationParams": evaluationParams,
                "evaluationLog": evaluationLog,
            }

    # Compute MAP and MAR
    AP = 0
    if evaluationParams["CONFIDENCES"]:
        AP = compute_ap(arrGlobalConfidences, arrGlobalMatches, numGlobalCareGt)

    methodRecall = 0 if numGlobalCareGt == 0 else float(matchedSum) / numGlobalCareGt
    methodPrecision = (
        0 if numGlobalCareDet == 0 else float(matchedSum) / numGlobalCareDet
    )
    methodHmean = (
        0
        if methodRecall + methodPrecision == 0
        else 2 * methodRecall * methodPrecision / (methodRecall + methodPrecision)
    )

    methodMetrics = {
        "precision": methodPrecision,
        "recall": methodRecall,
        "hmean": methodHmean,
        "AP": AP,
    }

    resDict = {
        "calculated": True,
        "Message": "",
        "method": methodMetrics,
        "per_sample": perSampleMetrics,
    }

    return resDict



import json
import os
from typing import List, Tuple

import numpy as np
from lanyocr.lanyocr_utils import download_model
from lanyocr.text_recognizer.paddleocr_base import PaddleOcrBase


class PaddleOcrChServer(PaddleOcrBase):
    def __init__(self, use_gpu: bool = True) -> None:
        print("Recognizer: PaddleOcrChServer")

        model_h = 32
        model_w = 320

        cur_dir = os.path.dirname(os.path.realpath(__file__))

        model_ignored_tokens = [0]
        accepted_characters = json.load(
            open(os.path.join(cur_dir, "dicts/paddleocr_en_dict.json"))
        )
        # accepted_characters = []
        model_characters = json.load(
            open(os.path.join(cur_dir, "dicts/ppocr_keys_v1_mod.json"))
        )
        # model_path = os.path.join(
        #     cur_dir, "../models/lanyocr-paddleocr-db-recognizer.onnx"
        # )
        model_path = download_model("lanyocr-paddleocr-db-recognizer.onnx")

        assert os.path.exists(model_path)

        super().__init__(
            use_gpu,
            model_ignored_tokens,
            model_characters,
            accepted_characters,
            model_path,
            model_w,
            model_h,
        )

        self.set_max_batch_size(8)



from typing import List
import cv2
import onnxruntime
import numpy as np
from lanyocr.angle_classifier import LanyOcrAngleClassifier
from lanyocr.lanyocr_utils import download_model


class PaddleOcrAngleClassifierv2(LanyOcrAngleClassifier):
    def __init__(self, use_gpu: bool = True):
        super().__init__(use_gpu)

        self.model_h = 48
        self.model_w = 192

        model_path = download_model("lanyocr-paddleocr-direction-classifier.onnx")

        # load model
        providers = ["CPUExecutionProvider"]
        if self.use_gpu:
            providers = ["CUDAExecutionProvider"]

        opts = onnxruntime.SessionOptions()
        opts.graph_optimization_level = (
            onnxruntime.GraphOptimizationLevel.ORT_ENABLE_EXTENDED
        )

        self.session = onnxruntime.InferenceSession(
            model_path, sess_options=opts, providers=providers
        )

    def infer(self, bgr_image, thresh=0.5) -> int:
        normed_img = self._resize_norm_img(bgr_image, [3, self.model_h, self.model_w])
        rotated_probs = self.session.run(None, {"x": [normed_img]})[0][0]

        if rotated_probs[1] > thresh:
            return 180

        return 0

    def infer_batch(self, bgr_images, thresh=0.5) -> List[int]:
        results: List[int] = []

        normed_imgs = []
        for img in bgr_images:
            normed_img = self._resize_norm_img(img, [3, self.model_h, self.model_w])
            normed_imgs.append(normed_img)

        rotated_probs = self.session.run(None, {"x": normed_imgs})[0]

        for prob in rotated_probs:
            rotated_angle = 180 if prob[1] > thresh else 0
            results.append(rotated_angle)

        return results

    # internal functions
    def _resize_norm_img(self, img, model_input_shape, padding=True):
        imgC, imgH, imgW = model_input_shape
        h = img.shape[0]
        w = img.shape[1]

        if h == imgH and w == imgW:
            resized_image = img
        elif not padding:
            resized_image = cv2.resize(
                img, (imgW, imgH), interpolation=cv2.INTER_LINEAR
            )
            resized_w = imgW
        else:
            ratio = imgH / float(h)
            _w = int(ratio * w)

            _img = cv2.resize(img, (_w, imgH))

            resized_image = np.zeros(shape=(imgH, imgW, 3), dtype=np.uint8)
            if _w > imgW:
                resized_image[:, :, :] = _img[:, :imgW, :]
            else:
                resized_image[:, :_w, :] = _img[:, :, :]

        resized_image = resized_image.astype("float32")
        if model_input_shape[0] == 1:
            resized_image = resized_image / 255
            resized_image = resized_image[np.newaxis, :]
        else:
            resized_image = resized_image.transpose((2, 0, 1)) / 255
        resized_image -= 0.5
        resized_image /= 0.5

        return resized_image



import os
from distutils.core import setup

import setuptools

readme = open("README.md", "r").read()

setup(
    name="lanyocr",  # How you named your package folder (MyLib)
    packages=[
        "lanyocr",
        "lanyocr.angle_classifier",
        "lanyocr.benchmarker",
        "lanyocr.text_merger",
        "lanyocr.text_detector",
        "lanyocr.text_recognizer",
        "lanyocr.text_recognizer.dicts",
    ],  # Chose the same as "name"
    version="0.1.1",  # Start with a small number and increase it with every change you make
    license="MIT",  # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description="An OCR library for Python",  # Give a short description about your library
    long_description=readme,
    long_description_content_type="text/markdown",
    author="JC1DA",  # Type in your name
    author_email="jc1da.3011@gmail.com",  # Type in your E-Mail
    url="https://github.com/JC1DA/lanyocr",  # Provide either the link to your github or to your website
    download_url="https://github.com/JC1DA/lanyocr",  # I explain this later on
    keywords=["ocr"],  # Keywords that define your package best
    install_requires=[
        "onnxruntime-gpu>=1.13",
        "numpy>=1.21",
        "opencv-python>=4.5",
        "pydantic>=1.10",
        "shapely",
        "pyclipper",
        "requests",
    ],  # I get to this in a second
    setup_requires=["wheel"],
    python_requires=">=3.0",
    classifiers=[
        "Development Status :: 3 - Alpha",  # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        "Intended Audience :: Developers",  # Define that your audience are developers
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",  # Again, pick a license
        "Programming Language :: Python :: 3.6",  # Specify which pyhton versions that you want to support
    ],
    include_package_data=True,
    data_files=[
        (
            "lanyocr/text_recognizer/dicts",
            [
                "lanyocr/text_recognizer/dicts/paddleocr_en_dict.json",
                "lanyocr/text_recognizer/dicts/paddleocr_french_dict.json",
                "lanyocr/text_recognizer/dicts/paddleocr_latin_dict.json",
                "lanyocr/text_recognizer/dicts/ppocr_keys_v1_mod.json",
            ],
        ),
        ("outputs", ["outputs/example1.jpg", "outputs/example2.jpg"]),
    ],
)
